/* S11N Example */
/* this is far away from reality, but demonstrates the most important features
 * of libs11n. Comments welcome.
 *
 * martin f. krafft <madduck@debian.org> 2004.05.13.2152
 *
 * Todo:
 *   - B::v does not deserialise, and when it does, it should maintain the
 *     original pointer relationship: (&a, &a, &b), not (&a, &b, &c).
 *   - BB does not yet deserialise. working on it.
 *   - show how to dump various objects into one file and load them from it
 *     again.
 *   - fix serialisation of B::aa
 */

#include <iostream>
#include <vector>
#include <list>
#include <map>
#include <s11n/s11nlite.h>
#include <s11n/algo.h> // s11n::free_list_entries
#include <s11n/map.h> // s11n::map namespace
#include <s11n/list.h> // s11n::list namespace

/* ========================================================================== */
/* standard simple class
 */

#define S11N_TYPE A
#define S11N_NAME "A"

struct A
{
  int i;
  std::string s;
  std::map<int, int> m;

  A() : i(5), s("martin"), m()
  {
    m[2] = 4;
    m[3] = 9;
    m[4] = 16;
  }
  virtual ~A() {}

  virtual bool
  operator()(s11nlite::node_type& dest) const
  {
    bool ret = true;
    dest.impl_class(S11N_NAME);
    dest.set("i", i);
    dest.set("s", s);
    ret &= (s11n::map::serialize_streamable_map(dest, "m", m) == m.size());
    return ret;
  }

  virtual bool
  operator()(s11nlite::node_type const& src)
  {
    bool ret = true;
    i = src.get("i", -1); ret &= (i != -1);
    s = src.get("s", "no string was set"); ret &= (s != "no string was set");
    m.clear();
    s11n::map::deserialize_streamable_map(src, "m", m);
    return ret;
  }
};

#include <s11n/reg_serializable.h>

std::ostream&
operator<<(std::ostream& out, A const& OBJ)
{
  return out << "(i=" << OBJ.i << ";s=\"" << OBJ.s
    << "\";m[3]=" << OBJ.m.find(3)->second << ')';
}

/* ========================================================================== */
/* standard derived class
 */

#define S11N_BASE_TYPE A
#define S11N_TYPE AA
#define S11N_NAME "AA"

struct AA : public A
{
  char c;

  AA() : A(), c('f') {}
  virtual ~AA() {}

  virtual bool
  operator()(s11nlite::node_type& dest) const
  {
    bool ret = A::operator()(dest);
    dest.impl_class(S11N_NAME);
    dest.set("c", c);
    return ret;
  }

  virtual bool
  operator()(s11nlite::node_type const& src)
  {
    bool ret = A::operator()(src);
    c = src.get<char>("c", '\0'); ret &= (c != '\0');
    return ret;
  }
};

#include <s11n/reg_serializable.h>

std::ostream&
operator<<(std::ostream& out, AA const& OBJ)
{
  return out << "(A=" << *reinterpret_cast<A const*>(&OBJ)
    << ";c=" << OBJ.c << ')';
}

/* ========================================================================== */
/* class with custom serialisation interface and containing pointers and
 * a list of pointers.
 *
 * CAVEAT: restoring the list of pointers does not (yet) work as expected. the
 * three items in the restored list are separate, even though 0 and 1 were the
 * same pointer.
 *
 * CAVEAT: i have not managed to get the pointer-storing of member aa to work.
 */

#define S11N_TYPE B
#define S11N_NAME "B"
#define S11N_SERIALIZE_FUNCTION serialise
#define S11N_DESERIALIZE_FUNCTION deserialise

struct B
{
  AA* aa;
  int x;
  std::vector<char> v;
  A a, b;
  std::list<A*> pv;

  B() : aa(new AA), x(7), v(), a(), b(), pv()
  {
    v.push_back('a');
    v.push_back('b');
    v.push_back('c');
    pv.push_back(&a);
    pv.push_back(&a);
    pv.push_back(&b);
  }
  virtual ~B()
  {
    if (aa) delete aa;
  }

  virtual bool
  serialise(s11nlite::node_type& dest) const
  {
    bool ret = true;
    dest.impl_class(S11N_NAME);
    //dest.set("aa", aa);
    dest.set("x", x);
    ret &= (s11n::list::serialize_streamable_list(dest, "v", v) == v.size());
    ret &= s11n::list::serialize_list(dest, "pv", pv);
    return ret;
  }

  virtual bool
  deserialise(s11nlite::node_type const& src)
  {
    bool ret = true;
    x = src.get("x", -1); ret &= (x != -1);
    //aa = src.get("aa", aa);
    v.clear();
    s11n::list::deserialize_streamable_list(src, "v", v);
    s11n::free_list_entries(pv);
    pv.clear();
    s11n::list::deserialize_list(src, "pv", pv);
    return ret;
  }
};

#include <s11n/reg_serializable.h>

std::ostream&
operator<<(std::ostream& out, B const& OBJ)
{
  out << "(aa=" << OBJ.aa;
  if (OBJ.aa) out << ";*aa=" << *OBJ.aa;
  out << ";x=" << OBJ.x << ";v=[";
  std::copy(OBJ.v.begin(), OBJ.v.end(),
      std::ostream_iterator<char>(std::cout, " "));
  out << "];pv=[";
  std::copy(OBJ.pv.begin(), OBJ.pv.end(),
      std::ostream_iterator<A*>(std::cout, " "));
  return out << "])";
}

/* ========================================================================== */
/* derived class using a proxy for serialisation
 */

#define S11N_BASE_TYPE B
#define S11N_TYPE BB
#define S11N_NAME "BB"
#define S11N_SERIALIZE_FUNCTOR BB_serialiser

class BB_serialiser; // friend decl

class BB : public B
{
private:
  char c;
  friend class BB_serialiser;
  friend std::ostream& operator<<(std::ostream&, BB const&);

public:
  BB() : B(), c('f') {}
  ~BB() {}
};

struct BB_serialiser
{
  virtual bool
  operator()(s11nlite::node_type& dest, BB const& OBJ) const
  {
    bool ret = OBJ.serialise(dest);
    dest.impl_class(S11N_NAME);
    dest.set("c", OBJ.c);
    return ret;
  }

  virtual bool
  operator()(s11nlite::node_type const& src, BB& OBJ) const
  {
    bool ret = OBJ.deserialise(src);
    OBJ.c = src.get<char>("c", '\0'); ret &= (OBJ.c != '\0');
    return ret;
  }
};

#include <s11n/reg_proxy.h>

std::ostream&
operator<<(std::ostream& out, BB const& OBJ)
{
  return out << "(B=" << *reinterpret_cast<B const*>(&OBJ)
    << ";c=" << OBJ.c << ')';
}

/* ========================================================================== */

#if 0
#define S11N_TYPE Ap
#define S11N_NAME "Ap"

struct Ap {
  A* a;

  Ap(A* const OBJ = NULL) : a(OBJ) {}
  Ap(Ap const& THAT) : a(THAT.a) {}
  Ap& operator=(Ap const& THAT) { this->a = THAT.a; return *this; }

  A* operator->() { return a; }
  A const* operator->() const { return a; }
  A& operator*() { return *a; }
  A const& operator*() const { return *a; }

  bool
  operator()(s11nlite::node_type& dest) const
  {
    dest.impl_class(S11N_NAME);
    dest.set("a", a);
    return true;
  }
  
  bool
  operator()(s11nlite::node_type const& src)
  {
    a = &src.get("a", *a);
    return true;
  }
};

#include <s11n/reg_serializable.h>

#define S11N_TYPE C
#define S11N_NAME "C"

std::ostream&
operator<<(std::ostream& out, Ap const& OBJ)
{
  return out << "(a=" << OBJ.a << ')';
}

struct C
{
  A a;
  Ap a1;
  Ap a2;

  C() : a(), a1(&a), a2(&a) {}

  bool
  operator()(s11nlite::node_type& dest) const
  {
    dest.impl_class(S11N_NAME);
    dest.set("a", a);
    dest.set("a1", a1);
    dest.set("a2", a2);
    return true;
  }
  
  bool
  operator()(s11nlite::node_type const& src)
  {
    a = src.get("a", a);
    a1 = src.get("a1", a1);
    a2 = src.get("a2", a2);
    return true;
  }
};

std::ostream&
operator<<(std::ostream& out, C const& OBJ)
{
  return out << "(a=" << OBJ.a << ";a1=" << OBJ.a1
    << ";a2=" << OBJ.a2 << ')';
}
#endif

/* ========================================================================== */

int main(void)
{
  A a; AA aa; B b; BB bb;

#define PRINT(x) std::cout << #x ":" << x << std::endl

  PRINT(a);
  PRINT(aa);
  PRINT(b);
  PRINT(bb);

#define SAVE(x) \
  std::cout << std::boolalpha << "saving " << #x "... " \
    << s11nlite::save(x, #x ".s11n") << std::endl

  SAVE(a);
  SAVE(aa);
  SAVE(b);
  SAVE(bb);

  std::cout << "press enter to continue...";
  std::cin.get();

#define LOAD(t, x) \
  t* x ## p = s11nlite::load_serializable<t>(#x ".s11n"); \
  if (!x ## p) std::cout << "reload of " #t " failed." << std::endl; \
  else PRINT(*x ## p)

  LOAD(A, a);
  LOAD(AA, aa);
  LOAD(B, b);
  LOAD(BB, bb);

  return 0;
}
